/*  Filename:	since.c
 *  Date:	10/16/87
 *  Author:	unknown (downloaded from Exec-PC, 10/15/87; the original
 *		was apparently uploaded to Exec-PC 8/18/87)
 *  Reviser:	Herb Robinson
 *
 *  Contents:	This is a program to list all the files on the current disk
 *		which have been touched since a given day (and, optionally,
 *		hour and minute). This version improves the original
 *		    1. by allowing the hour and minute to be specified
 *		       in addition to the month, day, and year, and
 *		       by providing defaults for missing arguments,
 *		    2. by improving the appearance of the printed output, and
 *		    3. by commenting and otherwise improving the appearance
 *		       of the source code.
 *
 *  Comments:	1. The original author used Microsoft C, version 3.0. I used
 *		   Microsoft C, version 4.0.
 *		2. Suggested improvements: extend this program so that it will
 *		    a. list files touched BETWEEN two given times,
 *		    b. accept a drive and/or a directory name, and
 *		    c. accept a filespec containing wild-card characters.
 *		3. Microsoft-specific functions used: intdos(), chdir().
 *		4. This program ignores the existence of February 29.
 */

#include <stdio.h>
#include <direct.h>
#include <dos.h>

struct dosdta
{
    unsigned char rsv[21];	    /* reserved for DOS */
    unsigned char attrib;	    /* attribute found	*/
    unsigned int  ftime;	    /* file's time      */
    unsigned int  fdate;	    /* file's date      */
    long	  fsize;	    /* file size	*/
    unsigned char fspec[13];	    /* file name.typ,0	*/
};

static unsigned int date = 0, time = 0;

static char *amonth[] =
{
    "January",
    "February",
    "March",
    "April",
    "May",
    "June",
    "July",
    "August",
    "September",
    "October",
    "November",
    "December"
};

static int dmonth[] = { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

static void findmod(char *), giveup(int);

/* main()
 */

void main(argc,argv)
int  argc;
char **argv;
{
    union REGS regs;
    char savedir[65];
    unsigned int mm, dd, yy, hh, nn;

    /* get the current date in case we have to take defaults */

    regs.h.ah = 0x2a;
    intdos(&regs, &regs);

    /* From 0 to 5 arguments must be given: the month, day, year, hour,
     * and minute. The defaults are as follows:
     *
     *	    Arguments	    Qualifying date and time
     *	    ---------	    ------------------------
     *		0	    since the start of today
     *		1	    since the start of the most recent given month
     *		2	    since the start of the most recent given day
     *		3	    since the start of the given day
     *		4	    since the start of the given hour
     *		5	    since the start of the given minute
     */

    if( argc > 6 )
	giveup(100);
    mm = (argc > 1) ? atoi(argv[1]) : regs.h.dh;
    dd = (argc > 2) ? atoi(argv[2]) : ((argc == 2) ? 1: regs.h.dl);
    yy = (argc > 3) ? atoi(argv[3]) : regs.x.cx;
    if ( mm > regs.h.dh	 ||  (mm == regs.h.dh && dd > regs.h.dl) )
	yy--;
    hh = (argc > 4) ? atoi(argv[4]) : 0;
    nn = (argc > 5) ? atoi(argv[5]) : 0;

    /* validate the arguments; refuse to proceed if any are faulty */

    if( (mm < 1)  ||  (mm > 12) )
    {
	fprintf(stderr,"Invalid month, must be >= 1 and <= 12.\n");
	giveup(4);
    }

    if( (( mm != 2)  &&	 (dmonth[mm-1] < dd)) || ((mm == 2) && (dd > 29)) )
    {
	fprintf(stderr,"Sorry, %s has %2d days.\n",amonth[mm-1],dmonth[mm-1]);
	giveup(4);
    }

    if( ((yy < 1980)  &&  (yy > 99)) || (yy < 80) )
    {
	fprintf(stderr,"Invalid year, must be >= 1980 or >79 and < 100.\n");
	giveup(4);
    }

    if( (hh < 0)  ||  (hh > 23) )
    {
	fprintf(stderr,"Invalid hour, must be >= 0 and <= 23.\n");
	giveup(4);
    }

    if( (nn < 0)  ||  (nn > 59) )
    {
	fprintf(stderr,"Invalid minute, must be >= 0 and <= 59.\n");
	giveup(4);
    }

    /* title the list */

    printf("\n");
    printf("Files since %02.2d-%02.2d-%d %02.2d:%02.2d\t", mm, dd, yy, hh, nn);

    /* pack the date and time as DOS does */

    yy = ((yy >= 1980) ? yy - 1980 : yy - 80) << 9;
    mm <<= 5;
    date = yy | mm | dd;
    hh <<= 11;
    nn <<= 5;
    time = hh | nn;
    printf("x\'%4.4x\' x\'%4.4x\'\n\n", date, time);

    /* print column headers */

    printf("Date\t\tTime\tFile\n----\t\t----\t----\n");

    /* remember the directory we started from */

    regs.x.ax = 0x4700;		     /* get current directory	 */
    regs.x.si = (short)&savedir[1];  /* where to put it		 */
    regs.x.dx = 0;		     /* default drive		 */
    intdos(&regs, &regs);	     /* do it			 */
    savedir[0] = '\\';		     /* start off with backslash */

    findmod("/");		     /* start at root directory */

    chdir(savedir);		     /* return to original directory */
}


static void giveup(n)
int n;
{
    fprintf(stderr, "Useage: since [mm [dd [yy [hh [mm]]]]]\n");
    exit(n);
}

/* findmod()
 *
 *	Traverse the tree rooted at the given subdirectory, and along
 *	the way, list the files that qualify.
 */

static void findmod(s)
char *s;
{
    static char allstr[] = "*.*";
    struct dosdta dta;
    char savedir[65];
    union REGS regs;

    chdir(s);			     /* change to specified directory */

    regs.x.si = (short)&savedir[1];  /* area to save current directory */
    regs.x.ax = 0x4700;		     /* get current directory function */
    regs.x.dx = 0;		     /* default drive		       */
    intdos(&regs, &regs);
    savedir[0] = (savedir[1] == '\0') ? '\0' : '\\';

    regs.x.dx = (short)&dta;	     /* set local DTA	 */
    regs.x.ax = 0x1a00;		     /* set DTA function */
    intdos(&regs, &regs);	     /* do it		 */

    regs.x.ax = 0x4e00;		     /* find first	       */
    regs.x.dx = (short)&allstr[0];   /* *.*		       */
    regs.x.cx = 0x10;		     /* include subdirectories */
    intdos(&regs, &regs);	     /* do it		       */

    /* for every file in this directory ... */

    while( regs.x.cflag == 0 )
    {
	if( (dta.attrib & 0x10)	 &&  dta.fspec[0] != '.' )
	{
	    /* if it's a subdirectory, recurse */

	    findmod(dta.fspec);
	    chdir("..");

	    regs.x.dx = (short)&dta;	     /* restore local DTA */
	    regs.x.ax = 0x1a00;		     /* set DTA function  */
	    intdos(&regs, &regs);	     /* do it		  */
	}
	else if ( dta.fspec[1] != '.' )
	{
	    /* else it's an ordinary file - list it if its date and
	     * time qualify; note the simplicity of the test - kudos
	     * to the authors of DOS
	     */

	    if(	   dta.fdate > date
	       ||  (dta.fdate == date && dta.ftime >= time) )
	    {
		printf("%02.2d-%02.2d-%d\t%02.2d:%02.2d\t%s\\",
		       (dta.fdate & 0x1e0) >> 5,
		       dta.fdate & 0x1f,
		       (dta.fdate >> 9) + 1980,
		       dta.ftime >> 11,
		       (dta.ftime >> 5) & 0x3f,
		       savedir);
		if ( dta.fspec[0] != '.' )
		    printf("%s", dta.fspec);
		printf("\n");
	    }
	}

	/* on to the next file */

	regs.x.ax = 0x4f00;	      /* find next */
	intdos(&regs, &regs);	      /* do it	   */
    }
}
